c$Id: cray_packer.f,v 1.1.1.1 1998/02/17 23:44:00 eaton Exp $
c***********************************************************************
c
c latest revision       Mar 09 1996.
c
c README                ***********  EVALUATION COPY  ***********
c
c                       Comments/questions/problems to dick valent
c                       valent@ncar.ucar.edu
c
c                       This file contains NCAR locally-developed PACKF
c                       routines, and their dependencies, for CRAY PVP 
c                       computers running CRAY floating-point arithmetic.
c
c                       Synopsis:
c
c                       This version is identical to the previous version
c                       (Dec 21 1994) except for routine packaf, which 
c                       has been modified to pack "near-zero" fields 
c                       (i.e. fields satisfying condition
c                           ABS((XMX-XMN)/PO2(NPK-1)) .LE. 1.0E-2455)
c                       In the 1994 version, routine packaf overflowed when 
c                       computing scale factor ABS(PO(2)/ABS(XMN-XMX)) for 
c                       these fields.
c
c purpose               To be used on CRAY PVP machines, to reduce data
c                       volume at the expense of accuracy, by packing 
c                       several values per Cray word.  See the "algorithm" 
c                       section below for details.
c
c usage                 call packaf (arn,nrn,apn,npk,amn,amx)
c                       call packbf (arn,nrn,apn,npk,amn,amx)
c                       call packcf (arn,nrn,apn,npk,amn,amx)
c                       call unpkaf (arn,nrn,apn,npk,amn,amx)
c                       call unpkbf (arn,nrn,apn,npk,amn,amx)
c                       call unpkcf (arn,nrn,apn,npk,amn,amx)
c
c                       There are three packing routines to
c                       choose from - packaf, packbf, and packcf.
c                       They differ only in the way in which
c                       amn and amx are computed.
c
c                       Packaf is the safest, but the slowest.
c                       amn and amx are computed by the routine
c                       itself.  Initially, amn is the actual
c                       minimum and amx is the actual maximum.
c                       then, if amn = amx, the values are
c                       adjusted outwards to avoid division by zero.
c                       further, if amn amx differ in sign,
c                       amx is adjusted outwards a little to
c                       ensure that zeroes in the original array
c                       will be returned by the unpacker as zeroes
c                       (mathematically, at least - hardware
c                       round-off makes it impossible to ensure
c                       the return of exact zeroes without
c                       seriously degrading the speed of the
c                       routine - however, the value returned should
c                       never be greater than 1.e-14*abs(amn) ).
c
c                       Packbf is the intermediate choice, faster
c                       but not as safe. amn and amx are computed
c                       by the routine itself.
c                       Initially, amx is set equal to the maximum
c                       absolute value and amn to the negative
c                       of that.  Then, amn and amx are adjusted
c                       as for packaf.  This routine should probably
c                       be used only for data which is known to have
c                       a distribution which is roughly symmetric
c                       around zero.
c
c                       Packcf is the fastest, but most dangerous,
c                       of the three. the user must supply the
c                       initial values of amn and amx.  These values
c                       are not adjusted in any way.
c                       No test is made to ensure that data
c                       values lie between amn and amx - errors may
c                       result.
c
c                       Routines 'unpkaf', 'unpkbf', and 'unpkcf'
c                       are used for the unpacking process.
c                       Actually, they are all synonymns, provided
c                       for the sake of consistency if and when
c                       planned extensions are added to this package.
c
c                       Note that, in any case, these routines
c                       must be used with care.  One wildly
c                       out-of-range datum can destroy the precision
c                       of the rest.  Moreover, data sets with
c                       unlike ranges should be packed separately.
c
c arguments             All of the packing and unpacking routines
c                       have the same calling sequence, with
c                       arguments arn, nrn, apn, npk, amn, and amx,
c                       the last two of which may be omitted
c                       except in calls to packcf.  The arguments
c                       may be described as follows:
c
c on input              arn(i),i=1,nrn
c
c                       an array of real numbers - input to the
c                       packers, output from the unpacker
c
c                       nrn     the length of arn - input
c
c                       apn(1)		minimum value - amn
c                       apn(2)		scale factor -
c					    (2**(64/npk)-1)/(amx-amn)
c
c
c                       npk
c
c                       packing density - number of data to be packed
c                       in one output word (2, 3, or 4) - input
c
c                       amn
c
c                       minimum - input to packcf, output from others
c                       (may be omitted from all except packcf calls)
c
c                       amx
c
c                       maximum - input to packcf, output from others
c                       (may be omitted from all except packcf calls)
c
c on output             apn(j) , j=3,4,...n
c
c                       the packed data - npk integers per word -
c                       n = 2+(nrn+npk-1)/npk - output from the
c                       packers, input to the unpacker
c
c special conditions    Near-zero fields packed by routine packaf
c                       are subject to a larger block relative error 
c                       than non-near zero fields.  The following table 
c                       indicates the growth of block relative error 
c                       for near zero fields where packing density is
c                       taken as 4:
c
c                       max block r.e.             data interval
c                       ______________________________________________
c                   
c                       0.152557E-04          [-1.0E-2454, +1.0E-2454]
c                       0.150864E-04          [-1.0E-2455, +1.0E-2455]
c                       0.150434E-04          [-1.0E-2456, +1.0E-2456]
c                       0.151844E-04          [-1.0E-2457, +1.0E-2457]
c                       0.146489E-04          [-1.0E-2458, +1.0E-2458]
c                       0.150809E-04          [-1.0E-2459, +1.0E-2459]
c                       0.151531E-04          [-1.0E-2460, +1.0E-2460]
c                       0.605122E-04          [-1.0E-2461, +1.0E-2461]
c                       0.961318E-03          [-1.0E-2462, +1.0E-2462]
c                       0.779736E-02          [-1.0E-2463, +1.0E-2463]
c                       0.624857E-01          [-1.0E-2464, +1.0E-2464]
c                       0.499992E+00          [-1.0E-2465, +1.0E-2465]
c
c                       Users who are troubled by this loss of accuracy 
c                       should consider screening the data in their
c                       application before packing.
c                       
c i/o                   none
c
c precision             see algorithm discussion below
c
c required library      none
c files
c
c language              fortran
c
c history               1979: Written by Vince Wayland of Cray Research, 
c                       Inc, and made available to NCAR in the same year.
c			1991: (September) Re-written entirely in fortran
c                       for use on CRAY-2 machines, and also XMP/YMP
c                       machines if desired.  The two versions produce
c                       different results on the XMP/YMP series, as
c                       explained above in the "purpose" section.
c
c algorithm             Given an array of real numbers
c                       (arn(i), i=1,nrn), a packing density
c                       npk (with value 2, 3, or 4), and real numbers
c                       amn and amx (the former less than or equal
c                       to any datum in arn and the latter greater
c                       than or equal to any datum in arn), one can
c                       reduce each element of arn to a positive
c                       integer less than or equal to
c                       2**(64/npk)-1 by means of the
c                       following formula:
c
c                       int(i) = ifix( (arn(i)-amn) *
c                       	 (2**(64/npk)-1) / (amx-amn) + .5 )
c
c                       the numbers amn and (2**(64/npk)-1) / (amx-amn)
c			may then be stored as apn(1) and apn(2), followed
c			by the integers (int(i),i=1,nrn), packed npk per
c			word in apn(3) through apn(2+(nrn+npk-1)/npk)).
c			The original (arn(i),i=1,nrn) may subsequently
c			be reconstructed from the contents of the packed
c			array apn by means of the following formula:
c
c                       arn(i) = amn + int(i) * (amx-amn) /
c                       	 (2**(64/npk)-1)
c
c			Note that the relative error of the reconstructed
c			values is a function of the packing density.
c			The error being 1. / (2**(64/npk)-1), i.e.
c			2.3 e-9, 4.8 e-7, 1.5 e-5 for npk 2, 3, 4
c			respectively.  However, these relative errors
c                       should not be taken seriously when the numbers 
c                       packed are close to zero, in terms of underflow.
c
c portability           not portable.
c
c timing                as of September, 1992 the following timings
c                       obtained on a Cray-YMP8/864.
c
c                       each time given is the time required
c                       to pack or unpack 100,000 numbers, in milliseconds.
c
c                       routine     npk=2      npk=3      npk=4
c                       -------     -----      -----      -----
c
c			packaf       4.97       5.03       5.11
c			packbf       3.60       3.67       3.76
c			packcf       2.06       2.12       2.21
c			unpkbf       1.79       1.46       1.30
c
c***********************************************************************
      subroutine packaf (arn,nrn,apn,npk,amn,amx)
      dimension  arn(nrn),apn(1),po2(3)
      LOGICAL    BIGSF
      COMMON    /PACKSF/ BIGSF
      DATA SAFRCP / 1.0e-2455 /
      data po2  /4294967295.,2097151.,65535. /
c
c check for certain errors in the arguments.
c
      noa=numarg()
      if (noa.lt.4 .or. noa.gt.6) go to 901
c
      if (nrn.le.0) go to 902
c
      if (npk.lt.2 .or. npk.gt.4) go to 903
c
      xmn=0577777777777777777777b
      xmx=1577777777777777777777b
c
      do 101 i = 1, nrn
	 xmn=min(xmn,arn(i))
	 xmx=max(xmx,arn(i))
  101 continue
      IF     ((XMX .EQ. XMN .and. XMX .gt. SAFRCP) 
     1 .OR.   (XMX .EQ. XMN .and. XMX .EQ. 0.0) 
     2 .OR.   (ABS((XMX-XMN)/PO2(NPK-1)) .GT. SAFRCP)) THEN
        BIGSF = .FALSE.
      ELSE
        BIGSF = .TRUE.
        IF (XMN .LT. 0.0 .AND. XMX .GT. 0.0) then
          XMN = -MAX(ABS(XMN), ABS(XMX))
          XMX = +MAX(ABS(XMN), ABS(XMX))
        ELSEIF (XMN .ge. 0.0) THEN
          XMN = 0.0
        ELSE
          XMX = 0.0
        ENDIF
 2415   IF (ABS((XMX-XMN)/PO2(NPK-1)) .LE. 1.0E-2465) THEN
          XMN=2.*XMN
          XMX=2.*XMX
          GO TO 2415
        ENDIF
      ENDIF
c
c adjust the values of xmn and xmx - make sure they are not equal.
c
  102 if (xmn.ne.xmx) go to 104
      xmn=xmn-.5*abs(xmn)
      xmx=xmx+.5*abs(xmx)
      if (xmn.ne.xmx) go to 104
      xmn=-1.
      xmx=+1.
c
c if xmn and xmx differ in sign, further adjust them so that zero will
c be a possible output from the unpacker.
c
  104 if (xmn*xmx.ge.0.) go to 105
      if (abs(xmn).lt.abs(xmx)) then
	  xmn=xmx-xmx*po2(npk-1)/float(ifix( xmx*po2(npk-1)/(xmx-xmn)))
      else
	  xmx=xmn-xmn*po2(npk-1)/float(ifix(-xmn*po2(npk-1)/(xmx-xmn)))
      endif
c
c return the computed values of xmn and xmx to the user, if appropriate.
c
  105 if (noa.ge.5) amn=xmn
      if (noa.ge.6) amx=xmx
c
c store the minimum and the scale factor in output array apn
c
      apn(1)=xmn
      apn(2)=po2(npk-1)/(xmx-xmn)
c
c call the appropriate packing routine.
c 
      if (npk .eq. 2) then
        call pack2f (arn, nrn, apn)
      else if (npk .eq. 3) then
        call pack3f (arn, nrn, apn)
      else
        call pack4f (arn, nrn, apn)
      endif
      return
c
c error exits.
c
  901 print 9001
      stop
c
  902 print 9002
      stop
c
  903 print 9003
      stop
c
c formats.
c
 9001 format (*0packaf - wrong number of arguments - stop*)
 9002 format (*0packaf - array length .le. 0 - stop*)
 9003 format (*0packaf - packing density .ne. 2, 3, or 4 - stop*)
c
      end
      subroutine packbf (arn,nrn,apn,npk,amn,amx)
      dimension arn(nrn),apn(1),po2(3)
      data po2 / 4294967295.,2097151.,65535. /
c
c check for certain errors in the arguments.
c
      noa=numarg()
      if (noa.lt.4 .or. noa.gt.6) go to 901
c
      if (nrn.le.0) go to 902
c
      if (npk.lt.2 .or. npk.gt.4) go to 903
c
      xmx=1577777777777777777777b
c
      do 101 i = 1, nrn
	 xmx=max(xmx,ABS(arn(i)))
  101 continue
c
  102 xmn=-xmx
c
c adjust the values of xmn and xmx - make sure they are not equal.
c
      if (xmn.ne.xmx) go to 103
      xmn=-1.
      xmx=+1.
c
c xmn and xmx differ in sign.  further adjust them so that zero will
c be a possible output from the unpacker.
c
  103 xmx=xmn-xmn*po2(npk-1)/float(ifix(-xmn*po2(npk-1)/(xmx-xmn)))
c
c return the computed values of xmn and xmx to the user, if appropriate.
c
  104 if (noa.ge.5) amn=xmn
      if (noa.ge.6) amx=xmx
c
c store the minimum and the scale factor in the output array.
c
      apn(1)=xmn
      apn(2)=po2(npk-1)/(xmx-xmn)
c
c call an appropriate routine to do the actual packing
c
      if (npk .eq. 2) then
	 call pack2f (arn, nrn, apn)
      else if (npk .eq. 3) then
	 call pack3f (arn, nrn, apn)
      else
	 call pack4f (arn, nrn, apn)
      endif
c
      return
c error exits.
c
  901 print 9001
      stop
c
  902 print 9002
      stop
c
  903 print 9003
      stop
c
c formats.
c
 9001 format (*0packbf - wrong number of arguments - stop*)
 9002 format (*0packbf - array length .le. 0 - stop*)
 9003 format (*0packbf - packing density .ne. 2, 3, or 4 - stop*)
c
      end
      subroutine packcf (arn,nrn,apn,npk,amn,amx)
c
c this is the subroutine packcf.
c
      dimension arn(nrn),apn(1),po2(3)
      data po2 / 4294967295.,2097151.,65535. /
c
c check for certain errors in the arguments.
c
      noa=numarg()
      if (noa.ne.6) go to 901
c
      if (nrn.le.0) go to 902
c
      if (npk.lt.2 .or. npk.gt.4) go to 903
c
      if (amn.ge.amx) go to 904
c
c store the minimum and the scale factor in the output array.
c
      apn(1)=amn
      apn(2)=po2(npk-1)/(amx-amn)
c
c call an appropriate routine to do the actual packing
c
      if (npk .eq. 2) then
	 call pack2f (arn, nrn, apn)
      else if (npk .eq. 3) then
	 call pack3f (arn, nrn, apn)
      else
	 call pack4f (arn, nrn, apn)
      endif
c
      return
c
c error exits.
c
  901 print 9001
      stop
c
  902 print 9002
      stop
c
  903 print 9003
      stop
c
  904 print 9004
      stop
c
c formats.
c
 9001 format (*0packcf - wrong number of arguments - stop*)
 9002 format (*0packcf - array length .le. 0 - stop*)
 9003 format (*0packcf - packing density .ne. 2, 3, or 4 - stop*)
 9004 format (*0packcf - minimum .ge. maximum - stop*)
c
      end
      subroutine pack2f (arn, nrn, apn)
c
c This routine packs 2 floats to the word
c
      dimension arn (nrn), apn (*)
      npc = nrn / 2
      xsf = apn(1)
      xsc = apn(2)
      do 101 j = 1, npc
	 i = j * 2 - 1
	 ih = (arn(i  ) - xsf) * xsc + 0.5
	 il = (arn(i+1) - xsf) * xsc + 0.5
	 apn(j+2) = or (shiftl (ih, 32), il)
  101 continue
c
c Pack up any left-overs
c
      if (npc * 2 .ne. nrn) then
	 j = npc + 1
	 i = j * 2 - 1
	 ih = (arn(i  ) - xsf) * xsc + 0.5
	 apn(j+2) = or (shiftl (ih, 32), 0)
      endif
      end
      subroutine pack3f (arn, nrn, apn)
c
c This routine packs 3 floats to the word
c
      dimension arn (nrn), apn (*)
      npc = nrn / 3
      xsf = apn(1)
      xsc = apn(2)
      do 101 j = 1, npc
	 i = j * 3 - 2
	 ih = (arn(i  ) - xsf) * xsc + 0.5
	 im = (arn(i+1) - xsf) * xsc + 0.5
	 il = (arn(i+2) - xsf) * xsc + 0.5
	 apn(j+2) = or (shiftl (ih, 42),
     ,                  or (shiftl (im, 21), il))
  101 continue
c
c Pack up any left-overs
c
      m = nrn - (npc * 3)
      if (m .ne. 0) then
	 j = npc + 1
	 i = j * 3 - 2
	 ih = (arn(i  ) - xsf) * xsc + 0.5
	 if (m .eq. 1) then
	    im = 0
	 else
	    im = (arn(i+1) - xsf) * xsc + 0.5
	 endif
	 apn(j+2) = or (shiftl (ih, 42),
     ,                  shiftl (im, 21))
	 endif
      end
      subroutine pack4f (arn, nrn, apn)
c
c This routine packs 4 floats to the word
c
      dimension arn (nrn), apn (*)
      npc = nrn / 4
      xsf = apn(1)
      xsc = apn(2)
      do 101 j = 1, npc
	 i = j * 4 - 3
	 ihh = (arn(i  ) - xsf) * xsc + 0.5
	 ihl = (arn(i+1) - xsf) * xsc + 0.5
	 ilh = (arn(i+2) - xsf) * xsc + 0.5
	 ill = (arn(i+3) - xsf) * xsc + 0.5
	 apn(j+2) = or (shiftl (ihh, 48), 
     ,		    or (shiftl (ihl, 32),
     ,		    or (shiftl (ilh, 16), ill)))
  101 continue
c
c Pack up any left-overs
c
      m = nrn - (npc * 4)
      if (m .gt. 0) then
	 j = npc + 1
	 i = j * 4 - 3
	 ihh = (arn(i  ) - xsf) * xsc + 0.5
	 ilh = 0
	 ihl = 0
	 goto (112, 111, 110) m
  110    ilh = (arn(i+2) - xsf) * xsc + 0.5
  111    ihl = (arn(i+1) - xsf) * xsc + 0.5
  112    continue
	 apn(j+2) = or (shiftl (ihh, 48), 
     ,		    or (shiftl (ihl, 32), shiftl (ilh, 16)))
      endif
      end
      subroutine unpack2f (arn, nrn, apn)
c
c This routine unpacks 2 floats from each word
c
      dimension arn (nrn), apn (*)
      npc = nrn / 2
      xsf = apn(1)
      xsc = 1.0 / apn(2)
      do 101 j = 1, npc
	 i = j * 2 - 1
	 ih =      shiftr (apn(j+2), 32)
	 il = and (        apn(j+2), 4294967295)
	 arn(i  ) = ih * xsc + xsf
	 arn(i+1) = il * xsc + xsf
  101 continue
c
c Unpack any left-overs
c
      m = nrn - (npc * 2)
      if (m .gt. 0) then
	 j = npc + 1
	 i = j * 2 - 1
	 ih = shiftr (apn(j+2), 32)
	 arn(i) = ih * xsc + xsf
      endif
      end
      subroutine unpack3f (arn, nrn, apn)
c
c This routine unpacks 3 floats from each word
c
      dimension arn (nrn), apn (*)
      npc = nrn / 3
      xsf = apn(1)
      xsc = 1.0 / apn(2)
      do 101 j = 1, npc
	 i = j * 3 - 2
	 ih =      shiftr (apn(j+2), 42)
	 im = and (shiftr (apn(j+2), 21), 2097151)
	 il = and (        apn(j+2),      2097151)
	 arn(i  ) = ih * xsc + xsf
	 arn(i+1) = im * xsc + xsf
	 arn(i+2) = il * xsc + xsf
  101 continue
c
c Unpack any left-overs
c
      m = nrn - (npc * 3)
      if (m .gt. 0) then
	 j = npc + 1
	 i = j * 3 - 2
	 ih = shiftr (apn(j+2), 42)
	 arn(i) = ih * xsc + xsf
	 if (m .gt. 1) then
	    im = and (shiftr (apn(j+2), 21), 2097151)
	    arn(i+1) = im * xsc + xsf
	 endif
      endif
      end
      subroutine unpack4f (arn, nrn, apn)
c
c This routine unpacks 4 floats from each word
c
      dimension arn (nrn), apn (*)
      npc = nrn / 4
      xsf = apn(1)
      xsc = 1.0 / apn(2)
      do 101 j = 1, npc
	 i = j * 4 - 3
	 ihh =      shiftr (apn(j+2), 48)
	 ihl = and (shiftr (apn(j+2), 32), 65535)
	 ilh = and (shiftr (apn(j+2), 16), 65535)
	 ill = and (        apn(j+2),      65535)
	 arn(i  ) = ihh * xsc + xsf
	 arn(i+1) = ihl * xsc + xsf
	 arn(i+2) = ilh * xsc + xsf
	 arn(i+3) = ill * xsc + xsf
  101 continue
c
c Unpack any left-overs
c
      m = nrn - (npc * 4)
      if (m .gt. 0) then
	 j = npc + 1
	 i = j * 4 - 3
	 ihh =      shiftr (apn(j+2), 48)
	 ihl = and (shiftr (apn(j+2), 32), 65535)
	 ilh = and (shiftr (apn(j+2), 16), 65535)
	 goto (112, 111, 110) m
  110    arn(i+2) = ilh * xsc + xsf
  111    arn(i+1) = ihl * xsc + xsf
  112    arn(i  ) = ihh * xsc + xsf
      endif
      end
      subroutine unpkbf (arn,nrn,apn,npk,amn,amx)
c
c this is the subroutine unpkbf.
c
      dimension arn(nrn),apn(1),msk(3)
      data msk / 4294967295 , 2097151 , 65535 /
c
c 'unpkaf', and 'unpkcf' are all synonyms for 'unpkbf'.
c
      entry unpkaf (arn,nrn,apn,npk,amn,amx)
      entry unpkcf (arn,nrn,apn,npk,amn,amx)
c
c check for errors in the input.
c
      noa=numarg()
      if (noa.lt.4 .or. noa.gt.6) go to 901
c
      if (nrn.le.0) go to 902
c
      if (npk.lt.2 .or. npk.gt.4) go to 903
c
c return the values of amn and amx, if appropriate.
c
      if (noa.ge.5) amn=apn(1)
      if (noa.ge.6) amx=apn(1)+float(msk(npk-1))/apn(2)
c
c call an appropriate routine to do the actual unpacking
c
      if (npk .eq. 2) then
	 call unpack2f (arn, nrn, apn)
      else if (npk .eq. 3) then
	 call unpack3f (arn, nrn, apn)
      else
	 call unpack4f (arn, nrn, apn)
      endif
c
  106 return
c
c error exits.
c
  901 print 9001
      stop
c
  902 print 9002
      stop
c
  903 print 9003
      stop
c
c formats.
c
 9001 format (*0unpkf(abc) - wrong number of arguments - stop*)
 9002 format (*0unpkf(abc) - array length .le. 0 - stop*)
 9003 format (*0unpkf(abc) - packing density .ne. 2, 3, or 4 - stop*)
c
c revision history -----------------------------------------------------
c
c mar 09     1996  Modify routine packaf to use safe large floating-point 
c                  arithmetic in calculating scale factor apn(2) when
c                  a "near-zero" field is encountered.
c
c dec 21     1994  Fix loop 101 of routine packbf to calculate xmx as
c                  max ABS ARN(I) instead of max ARN(I), for agreement 
c                  with Fortran-CAL packb routine.  
c
c august     1993  Added comments about inadvisability of short list 
c                  calls.
c
c september, 1991  Rewrote these routines in Fortran, for use on CRAY-2
c                  machines.  This version does not produce results which 
c                  are identical to results produced by the mixed Fortran
c                  /Cal routines used on the YMP machines. The reason is
c                  stated above in the "purpose" section of the package 
c                  documentation.
c ----------------------------------------------------------------------
      end
